// This program illustrates implement a worklist algorithm, using multiple
// processes and one POSIX message queue.
#define _XOPEN_SOURCE 600

#include "buf.h"
#include "kmp.h"

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <ftw.h>
#include <mqueue.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#define UNUSED(x) (void)(x)


int workers_count = 1; // default value

char * todo_queue_name = "/pgrep-todo";
mqd_t todo_queue;

void usage(const char * name) {
  fprintf(stderr, "usage: %s [-w workers] needle\n", name);
  fprintf(stderr, "  where 'workers' is one of 1,2,...,1024\n");
  exit (EXIT_FAILURE);
}


int search_path(
    const char * fpath,
    const struct stat * sb,
    int typeflag,
    struct FTW * ftwbuf
) {
  UNUSED(sb); UNUSED(ftwbuf);
  if (typeflag != FTW_F) return 0;
  // TODO: grow queue msg size if filename is too long
  // (Don't forget about the buffer on the receiving side, in do_grep).
  int command_size = 6 + strlen(fpath);
  char * command = malloc(command_size);
  if (command == NULL) {
    perror("malloc");
    killpg(getpgrp(), SIGKILL);
  }
  sprintf(command, "grep %s", fpath);
  int error = mq_send(todo_queue, command, command_size, 31);
  if (error == -1) {
    fprintf(stderr, "command %s\n  ", command);
    perror("mq_send");
    killpg(getpgrp(), SIGKILL);
  }
  if (0) printf("SENT %s\n", command);
  free(command);
  return 0;
}

void do_grep() {
  char * buffer = NULL;
  int buffer_size = 0;
  while (1) {
    int error = mq_receive(todo_queue, buffer, buffer_size, NULL);
    if (error == -1) {
      if (errno == EMSGSIZE) {
        struct mq_attr queue_attributes;
        error = mq_getattr(todo_queue, &queue_attributes);
        if (error == -1) {
          perror("mq_getattr");
          killpg(getpgrp(), SIGKILL);
        }
        free(buffer);
        buffer_size = queue_attributes.mq_msgsize;
        if (0) printf("growing to %d\n", buffer_size);
        buffer = malloc(buffer_size);
        if (buffer == NULL) {
          perror("malloc");
          killpg(getpgrp(), SIGKILL);
        }
        continue;
      } else {
        perror("mq_receive");
        killpg(getpgrp(), SIGKILL);
      }
    }
    buffer[buffer_size - 1] = 0;
    if (0) printf("RECEIVED %s\n", buffer);
    if (strncmp(buffer, "stop", 4) == 0) {
      mq_close(todo_queue);
      exit (EXIT_SUCCESS);
    } else if (strncmp(buffer, "grep ", 5) == 0) {
      if (!buf_open(buffer + 5)) continue;
      if (kmp_find(buf_get)) printf("%s\n", buffer + 5);
      buf_close();
    } else {
      fprintf(stderr, "ignoring unknown command: %s\n", buffer);
    }
  }
}

// TODO: handle only one needle?
void hire_workers() {
  mq_unlink(todo_queue_name);
  todo_queue = mq_open(todo_queue_name, O_WRONLY | O_CREAT, S_IRWXU, NULL);
  if (todo_queue == -1) {
    perror("mq_open main todo");
    exit (EXIT_FAILURE);
  }
  setpgid(0,0);
  // TODO: register handler for sigkill, to exit with EXIT_FAILURE
  for (int i = 0; i < workers_count; ++i) {
    pid_t p = fork();
    if (p == -1) {
      perror("fork");
      killpg(getpgrp(), SIGKILL);
    } else if (p == 0) { // in child
      todo_queue = mq_open(todo_queue_name, O_RDONLY);
      if (todo_queue == -1) {
        perror("mq_open worker todo");
        killpg(getpgrp(), SIGKILL);
      }
      do_grep();
      assert (0);
    } else { /* in parent, just continue */ }
  }
}

void fire_workers() {
  for (int i = 0; i < workers_count; ++i) {
    int error = mq_send(todo_queue, "stop", 5, 0);
    if (error == -1) {
      perror("mq_send stop");
      killpg(getpgrp(), SIGKILL);
    }
  }
  while (wait(NULL) != -1);
  if (errno != ECHILD) perror("wait");
  mq_close(todo_queue);
  mq_unlink(todo_queue_name);
}


int main(int argc, char * argv[]) {
  int option;
  while ((option = getopt(argc, argv, "w:")) != -1) {
    if (option == 'w') {
      if (sscanf(optarg, "%d", &workers_count) != 1) usage(argv[0]);
    } else usage(argv[0]);
    if (! (0 < workers_count && workers_count <= 1024)) {
      fprintf(stderr, "the number of workers must be in [1,1024]\n");
      usage(argv[0]);
    }
  }
  if (optind != argc - 1) usage(argv[0]);
  kmp_precompute(argv[optind]);
  hire_workers();
  int error = nftw(".", search_path, 1<<10, FTW_PHYS);
  if (error != 0) exit (EXIT_FAILURE);
  kmp_free();
  fire_workers();
  exit (EXIT_SUCCESS);
}
